"""Asynchronous client for the PVOutput API."""

# pylint: disable=protected-access
import socket
from datetime import UTC, date, datetime, time
from unittest.mock import patch

import aiohttp
import pytest
from aioresponses import aioresponses
from syrupy.assertion import SnapshotAssertion
from yarl import URL

from pvo import PVOutput
from pvo.exceptions import (
    PVOutputAuthenticationError,
    PVOutputConnectionError,
    PVOutputError,
    PVOutputNoDataError,
)


async def test_request() -> None:
    """Test response is handled correctly."""
    with aioresponses() as mocked:
        mocked.get(
            "https://pvoutput.org/service/r2/test",
            status=200,
            body="ok",
            content_type="text/plain",
        )
        async with aiohttp.ClientSession() as session:
            pvoutput = PVOutput(api_key="fake", system_id=12345, session=session)
            response = await pvoutput._request("test")
            assert response == "ok"
            await pvoutput.close()


async def test_internal_session() -> None:
    """Test response is handled correctly."""
    with aioresponses() as mocked:
        mocked.get(
            "https://pvoutput.org/service/r2/test",
            status=200,
            body="ok",
            content_type="text/plain",
        )
        async with PVOutput(api_key="fake", system_id=12345) as pvoutput:
            response = await pvoutput._request("test")
            assert response == "ok"


async def test_post_request() -> None:
    """Test post requests are handled correctly."""
    with aioresponses() as mocked:
        mocked.post(
            "https://pvoutput.org/service/r2/test",
            status=200,
            body="ok",
            content_type="text/plain",
        )
        async with aiohttp.ClientSession() as session:
            pvoutput = PVOutput(api_key="fake", system_id=12345, session=session)
            response = await pvoutput._request(
                "test",
                method=aiohttp.hdrs.METH_POST,
                data={},
            )
            assert response == "ok"


async def test_timeout() -> None:
    """Test request timeout from the API."""
    with aioresponses() as mocked:
        mocked.get(
            "https://pvoutput.org/service/r2/test",
            exception=TimeoutError(),
        )
        async with aiohttp.ClientSession() as session:
            pvoutput = PVOutput(
                api_key="fake",
                system_id=12345,
                session=session,
                request_timeout=0.1,
            )
            with pytest.raises(PVOutputConnectionError):
                assert await pvoutput._request("test")


async def test_http_error400() -> None:
    """Test HTTP 404 response handling."""
    with aioresponses() as mocked:
        mocked.get(
            "https://pvoutput.org/service/r2/test",
            status=404,
            body="OMG PUPPIES!",
            content_type="text/plain",
        )
        async with aiohttp.ClientSession() as session:
            pvoutput = PVOutput(api_key="fake", system_id=12345, session=session)
            with pytest.raises(PVOutputError):
                assert await pvoutput._request("test")


async def test_http_error401() -> None:
    """Test HTTP 401 response handling."""
    with aioresponses() as mocked:
        mocked.get(
            "https://pvoutput.org/service/r2/test",
            status=401,
            body="Access denied!",
            content_type="text/plain",
        )
        async with aiohttp.ClientSession() as session:
            pvoutput = PVOutput(api_key="fake", system_id=12345, session=session)
            with pytest.raises(PVOutputAuthenticationError):
                assert await pvoutput._request("test")


async def test_communication_error() -> None:
    """Test communication error handling."""
    async with aiohttp.ClientSession() as session:
        pvoutput = PVOutput(api_key="fake", system_id=12345, session=session)
        with (
            patch.object(
                session,
                "request",
                side_effect=socket.gaierror,
            ),
            pytest.raises(PVOutputConnectionError),
        ):
            assert await pvoutput._request("test")


async def test_get_status(snapshot: SnapshotAssertion) -> None:
    """Test get status handling."""
    with aioresponses() as mocked:
        mocked.get(
            "https://pvoutput.org/service/r2/getstatus.jsp",
            status=200,
            body="20211222,18:00,3636,0,NaN,NaN,NaN,21.2,220.1",
            content_type="text/plain",
        )
        async with aiohttp.ClientSession() as session:
            pvoutput = PVOutput(api_key="fake", system_id=12345, session=session)
            status = await pvoutput.status()

    assert status.reported_date == date(2021, 12, 22)
    assert status.reported_time == time(18, 0)
    assert status.reported_datetime == datetime(
        2021,
        12,
        22,
        18,
        0,
        tzinfo=UTC,
    )
    assert status.energy_consumption is None
    assert status.energy_generation == 3636
    assert status.normalized_output is None
    assert status.power_consumption is None
    assert status.power_generation == 0
    assert status.temperature == 21.2
    assert status.voltage == 220.1

    assert status.to_dict() == snapshot


async def test_get_status_no_data() -> None:
    """Test PVOutput status without data is handled."""
    with aioresponses() as mocked:
        mocked.get(
            "https://pvoutput.org/service/r2/getstatus.jsp",
            status=400,
            body="Bad Request!",
            content_type="text/plain",
        )
        async with aiohttp.ClientSession() as session:
            pvoutput = PVOutput(api_key="fake", system_id=12345, session=session)
            with pytest.raises(PVOutputNoDataError):
                await pvoutput.status()


async def test_get_system(snapshot: SnapshotAssertion) -> None:
    """Test get system handling."""
    with aioresponses() as mocked:
        mocked.get(
            "https://pvoutput.org/service/r2/getsystem.jsp",
            status=200,
            body=(
                "Frenck,5015,CO1,17,295,JA solar JAM-300,1,5000,"
                "SolarEdge SE5000H,S,20.0,Low,20180622,51.1234,6.1234,5;;0"
            ),
            content_type="text/plain",
        )
        async with aiohttp.ClientSession() as session:
            pvoutput = PVOutput(api_key="fake", system_id=12345, session=session)
            system = await pvoutput.system()

    assert system.array_tilt == 20.0
    assert system.install_date == date(2018, 6, 22)
    assert system.inverter_brand == "SolarEdge SE5000H"
    assert system.inverter_power == 5000
    assert system.inverters == 1
    assert system.latitude == 51.1234
    assert system.longitude == 6.1234
    assert system.orientation == "S"
    assert system.panel_brand == "JA solar JAM-300"
    assert system.panel_power == 295
    assert system.panels == 17
    assert system.shade == "Low"
    assert system.status_interval == 5
    assert system.system_name == "Frenck"
    assert system.system_size == 5015
    assert system.zipcode == "CO1"

    assert system.to_dict() == snapshot


async def test_add_status() -> None:
    """Test adding a status update."""
    with aioresponses() as mocked:
        mocked.post(
            "https://pvoutput.org/service/r2/addstatus.jsp",
            status=200,
            body="OK 200: Added Status",
            content_type="text/plain",
        )
        async with aiohttp.ClientSession() as session:
            pvoutput = PVOutput(api_key="fake", system_id=12345, session=session)
            await pvoutput.add_status(
                reported_date=date(2021, 12, 22),
                reported_time=time(14, 30),
                energy_generation=3636,
                power_generation=500,
                energy_consumption=1200,
                power_consumption=350,
                temperature=21.2,
                voltage=220.1,
            )

        assert mocked.requests is not None
        data = mocked.requests[
            ("POST", URL("https://pvoutput.org/service/r2/addstatus.jsp"))
        ][0].kwargs["data"]
        assert data["d"] == "20211222"
        assert data["t"] == "14:30"
        assert data["v1"] == "3636"
        assert data["v2"] == "500"
        assert data["v3"] == "1200"
        assert data["v4"] == "350"
        assert data["v5"] == "21.2"
        assert data["v6"] == "220.1"
        assert "c1" not in data
        assert "n" not in data


async def test_add_status_with_flags() -> None:
    """Test adding a status update with cumulative and net flags."""
    with aioresponses() as mocked:
        mocked.post(
            "https://pvoutput.org/service/r2/addstatus.jsp",
            status=200,
            body="OK 200: Added Status",
            content_type="text/plain",
        )
        async with aiohttp.ClientSession() as session:
            pvoutput = PVOutput(api_key="fake", system_id=12345, session=session)
            await pvoutput.add_status(
                reported_date=date(2021, 12, 22),
                reported_time=time(14, 30),
                energy_generation=50000,
                energy_consumption=12000,
                cumulative=True,
                net=True,
            )

        assert mocked.requests is not None
        data = mocked.requests[
            ("POST", URL("https://pvoutput.org/service/r2/addstatus.jsp"))
        ][0].kwargs["data"]
        assert data["c1"] == "1"
        assert data["n"] == "1"
        assert data["v1"] == "50000"
        assert data["v3"] == "12000"


async def test_add_status_defaults_date_time() -> None:
    """Test adding a status defaults to current date and time."""
    with aioresponses() as mocked:
        mocked.post(
            "https://pvoutput.org/service/r2/addstatus.jsp",
            status=200,
            body="OK 200: Added Status",
            content_type="text/plain",
        )
        async with aiohttp.ClientSession() as session:
            pvoutput = PVOutput(api_key="fake", system_id=12345, session=session)
            await pvoutput.add_status(power_generation=500)

        assert mocked.requests is not None
        data = mocked.requests[
            ("POST", URL("https://pvoutput.org/service/r2/addstatus.jsp"))
        ][0].kwargs["data"]
        assert "d" in data
        assert "t" in data
        assert data["v2"] == "500"


async def test_get_system_empty_install_date() -> None:
    """Test get system handling with an empty install date."""
    with aioresponses() as mocked:
        mocked.get(
            "https://pvoutput.org/service/r2/getsystem.jsp",
            status=200,
            body=(
                "Frenck,5015,1234,17,295,JA solar JAM-300,1,5000,"
                "SolarEdge SE5000H,S,20.0,Low,,51.1234,6.1234,5;;0"
            ),
            content_type="text/plain",
        )
        async with aiohttp.ClientSession() as session:
            pvoutput = PVOutput(api_key="fake", system_id=12345, session=session)
            system = await pvoutput.system()

    assert system.install_date is None



"""Asynchronous client for the PVOutput API."""

from __future__ import annotations

import asyncio
import socket
from dataclasses import dataclass
from datetime import UTC, date, datetime, time
from typing import Any, Self

from aiohttp.client import ClientError, ClientResponseError, ClientSession
from aiohttp.hdrs import METH_GET, METH_POST
from yarl import URL

from .exceptions import (
    PVOutputAuthenticationError,
    PVOutputConnectionError,
    PVOutputError,
    PVOutputNoDataError,
)
from .models import Status, System


@dataclass
class PVOutput:
    """Main class for handling connections with the PVOutput API."""

    api_key: str
    system_id: int

    request_timeout: float = 8.0
    session: ClientSession | None = None

    _close_session: bool = False

    async def _request(
        self,
        uri: str,
        *,
        method: str = METH_GET,
        data: dict[str, Any] | None = None,
    ) -> str:
        """Handle a request to the PVOutput API.

        A generic method for sending/handling HTTP requests done against
        the PVOutput API.

        Args:
        ----
            uri: Request URI, without '/service/r2/'.
            method: HTTP method to use.
            data: Dictionary of parameters to send to the PVOutput API.

        Returns:
        -------
            The response body from the PVOutput API.

        Raises:
        ------
            PVOutputAuthenticationError: If the API key is invalid.
            PVOutputConnectionError: An error occurred while communicating with
                the PVOutput API.
            PVOutputError: Received an unexpected response from the PVOutput
                API.

        """
        url = URL("https://pvoutput.org/service/r2/").join(URL(uri))

        headers = {
            "Accept": "application/json",
            "X-Pvoutput-Apikey": self.api_key,
            "X-Pvoutput-SystemId": str(self.system_id),
        }

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        try:
            async with asyncio.timeout(self.request_timeout):
                response = await self.session.request(
                    method,
                    url,
                    data=data,
                    headers=headers,
                )
                response.raise_for_status()
        except TimeoutError as exception:
            msg = "Timeout occurred while connecting to the PVOutput API"
            raise PVOutputConnectionError(msg) from exception
        except ClientResponseError as exception:
            if exception.status == 400 and uri.startswith("getstatus.jsp"):
                msg = "PVOutput has no status data available for this system"
                raise PVOutputNoDataError(msg) from exception
            if exception.status in [401, 403]:
                msg = "Authentication to the PVOutput API failed"
                raise PVOutputAuthenticationError(msg) from exception
            msg = "Error occurred while connecting to the PVOutput API"
            raise PVOutputError(msg) from exception
        except (ClientError, socket.gaierror) as exception:
            msg = "Error occurred while communicating with the PVOutput API"
            raise PVOutputConnectionError(msg) from exception

        return await response.text()

    async def status(self) -> Status:
        """Retrieve system status information and live output data.

        Returns
        -------
            A PVOutput Status object.

        """
        data = await self._request("getstatus.jsp")
        return Status.from_dict(
            dict(
                zip(
                    [
                        "reported_date",
                        "reported_time",
                        "energy_generation",
                        "power_generation",
                        "energy_consumption",
                        "power_consumption",
                        "normalized_output",
                        "temperature",
                        "voltage",
                    ],
                    data.split(","),
                    strict=True,
                )
            ),
        )

    async def system(self) -> System:
        """Retrieve system information.

        Returns
        -------
            A PVOutput System object.

        """
        data = await self._request("getsystem.jsp")
        return System.from_dict(
            dict(
                zip(
                    [
                        "system_name",
                        "system_size",
                        "zipcode",
                        "panels",
                        "panel_power",
                        "panel_brand",
                        "inverters",
                        "inverter_power",
                        "inverter_brand",
                        "orientation",
                        "array_tilt",
                        "shade",
                        "install_date",
                        "latitude",
                        "longitude",
                        "status_interval",
                    ],
                    data.partition(";")[0].split(","),
                    strict=True,
                )
            ),
        )

    # pylint: disable-next=too-many-arguments
    async def add_status(  # noqa: PLR0913
        self,
        *,
        energy_generation: int | None = None,
        power_generation: int | None = None,
        energy_consumption: int | None = None,
        power_consumption: int | None = None,
        temperature: float | None = None,
        voltage: float | None = None,
        cumulative: bool = False,
        net: bool = False,
        reported_date: date | None = None,
        reported_time: time | None = None,
    ) -> None:
        """Upload a live status update to PVOutput.

        Args:
        ----
            energy_generation: Energy generation in watt hours (Wh).
            power_generation: Power generation in watts (W).
            energy_consumption: Energy consumption in watt hours (Wh).
            power_consumption: Power consumption in watts (W).
            temperature: Temperature in celsius (C).
            voltage: Voltage in volts (V).
            cumulative: Whether energy values are lifetime cumulative.
            net: Whether values are net import/export.
            reported_date: Date of the status. Defaults to today.
            reported_time: Time of the status. Defaults to now.

        """
        now = datetime.now(tz=UTC)
        data: dict[str, str] = {
            "d": (reported_date or now.date()).strftime("%Y%m%d"),
            "t": (reported_time or now.time()).strftime("%H:%M"),
        }

        if energy_generation is not None:
            data["v1"] = str(energy_generation)
        if power_generation is not None:
            data["v2"] = str(power_generation)
        if energy_consumption is not None:
            data["v3"] = str(energy_consumption)
        if power_consumption is not None:
            data["v4"] = str(power_consumption)
        if temperature is not None:
            data["v5"] = str(temperature)
        if voltage is not None:
            data["v6"] = str(voltage)
        if cumulative:
            data["c1"] = "1"
        if net:
            data["n"] = "1"

        await self._request("addstatus.jsp", method=METH_POST, data=data)

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The PVOutput object.

        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exception info.

        """
        await self.close()
